#include <stdio.h>
#include <string.h>
#include <time.h>
#include "vec.h"
#include "pbc.h"
#include "random.h"
#include "gmx_random.h"
#include "physics.h"
#include "network.h"
#include "smalloc.h"
#include "mdrun.h"
#include "txtdump.h"
#include "macros.h"
#include "nonbonded.h"
#include "nrnb.h"

#define VW12(ti,tj) (C12(fr->nbfp, fr->ntype, (ti), (tj)))
#define VW6(ti,tj)   (C6(fr->nbfp, fr->ntype, (ti), (tj)))
#define LJSA(ai,aj) (VW12(mdatoms->typeA[(ai)], mdatoms->typeA[(aj)]))
#define LJSB(ai,aj)  (VW6(mdatoms->typeA[(ai)], mdatoms->typeA[(aj)]))
#define KB 8.314510e-3
#define KMAX 20
#define ALMOST_ZERO 1e-30
#define ALIVE(m) ((mc->gtype == top->atoms.atom[mc->first[LIPB][(m)]].type)?LIPA:LIPB)
#define RESNAME(g) (*(top->atoms.grpname[top->atoms.grps[egcTC].nm_ind[(g)]]))
#define RANMOL(mc) ((int)(rando(&((mc)->seed))*(mc)->nmol))
#define TIMER 0

static real rc2;
FILE *mclog;
#if TIMER
enum {T_LJ, T_BSEARCH, T_BOND, T_RANX, T_NRDF, T_RANV, T_MUTATE, 
      T_OLD, T_NEW, T_ACCREJ, T_DEBONDS, T_NR};
clock_t stopwatch[T_NR];
#endif
mc_ftypes[NFNB] = {F_SHAKE,F_ANGLES,F_PDIHS,F_IDIHS,F_RBDIHS,F_LJ14};

int countB (t_mc *mc, t_topology *top)
{
  int nB=0,m;
  for (m=0; m<mc->nmol; m++) nB += ALIVE(m);
  return nB;
}

/* von Neumann uniform variate on surface of a sphere */
void ran_x (real length, gsl_rng *rng, rvec x)
{
  int d; real r2=1,r,t;
#if TIMER
  clock_t start,stop;
  start = clock();
#endif
  while (r2 > 0.25) { /* avg 5.7 tries */
//    for (d=0; d<DIM; d++) x[d] = rando(seed) - 0.5;	/* 6	*/
    for (d=0; d<DIM; d++) x[d] = gsl_rng_uniform(rng) - 0.5;	/* 6	*/
    r2 = norm2(x);					/* 5	*/
  }						/* 63	*/
  r = sqrt(r2);					/* 10	*/
  for (d=0; d<DIM; d++) x[d] *= (length/r);	/* 33	*/
#if TIMER
  stop = clock();
  stopwatch[T_RANX] += stop - start;
#endif
}						/* TOTAL 109	*/

/* Adapted from "low_mspeed" in random.c but with gsl functions */
void gsl_mspeed(real tempi,int nrdf,int nat,atom_id a[],
                t_atoms *atoms,rvec v[],gsl_rng *rng)
{
  int  i,j,m;
  real boltz,sd;
  real ekin,temp,mass,scal;
//  gmx_rng_t gmx_rng;
//  gmx_rng = gmx_rng_init(5555);

  boltz=BOLTZ*tempi;
  ekin=0.0;
  for (i=0; (i<nat); i++) {
    j=a[i];
    mass=atoms->atom[j].m;
    if (mass > 0) {
      sd=sqrt(boltz/mass);
      for (m=0; (m<DIM); m++) {
        v[j][m]=gsl_ran_gaussian(rng,sd);
//        v[j][m]=sd*gmx_rng_gaussian_real(gmx_rng);
        ekin+=0.5*mass*v[j][m]*v[j][m];
      }
    }
  }
  temp=(2.0*ekin)/(nrdf*BOLTZ);
  if (temp > 0) {
    scal=sqrt(tempi/temp);
    for(i=0; (i<nat); i++)
      for(m=0; (m<DIM); m++)
        v[a[i]][m]*=scal;
  }
}

void make_vel(int new,int m,real T,t_mc *mc,t_atoms *atoms,rvec v[],t_inputrec *ir)
{
  int nrdf[1],nt1,nt2,i,ai;
  t_block b;
  real tempi[1];
#if TIMER
  clock_t start,stop;
  start = clock();
#endif
//  nt1 = max(0,mc->t[T1].nr[new]-mc->t[T1].nr[!new]);
//  nt2 = max(0,mc->t[T2].nr[new]-mc->t[T2].nr[!new]);
  nt1 = mc->t[T1].nr[new] - mc->t[T1].nsame;
  nt2 = mc->t[T2].nr[new] - mc->t[T2].nsame;
  b.nra = nt1+nt2;
  if (b.nra) {
    b.nr=1;
    snew(b.a,b.nra);
    snew(b.index,b.nr+1);
    b.index[0] = 0;
    b.index[1] = b.nra;
    nrdf[0] = 2*b.nra;
    ai = mc->first[new][m] + mc->t[T1].node[new] + 1;
    for (i=0; i<b.nra; i++,ai++) {
      if (i==nt1)
        ai = mc->first[new][m] + mc->t[T2].node[new] + 1;
      b.a[i]=ai;
    }
    tempi[0]=T;
//    grp_maxwell(&b,tempi,nrdf,mc->seed,atoms,v);
    grp_maxwell(&b,tempi,nrdf,5555,atoms,v);
    sfree(b.a);
    sfree(b.index);
  }
#if TIMER
  stop = clock();
  stopwatch[T_RANV] += stop - start;
#endif
}

void write_ndxtraj(char *outf, int step, t_atoms *atoms)
/* write out real molecule resid's */
{
  FILE *out;
  int  i,j,res,last_res=-1;
  static bool bFirst=TRUE;

  if (bFirst) {
    out=ffopen(outf,"w");
    bFirst=FALSE;
  } else
    out=ffopen(outf,"a");
  fprintf(out,"[ step %d ]\n",step);
  for(i=0,j=0; i<atoms->nr; i++) {
    res = atoms->atom[i].resnr;
    if ((atoms->atom[i].type == atoms->atom[i].typeB) &&
        (res != last_res)) {
      fprintf(out,"%5d ",res+1);
      if ((j % 12) == 11)
        fprintf(out,"\n");
      j++;
    }
    last_res = res;
  }
  if ((j % 12) != 12)
    fprintf(out,"\n");
  fclose(out);
}

real elj_i_j(int i, int j, rvec x[], t_pbc *pbc, 
             t_forcerec *fr, t_mdatoms *mdatoms, t_nrnb *nrnb)
{
  int nlj=0;
  rvec dx;
  real r,e=0;
  pbc_dx(pbc,x[i],x[j],dx);	/* 6	*/
  r = norm2(dx);		/* 5	*/
  if (r<=rc2) {
    r = 1.0/(r*r*r);			/* 12	*/
    e = LJSA(i,j)*r*r - LJSB(i,j)*r;	/* 4	*/
    nlj++;
  }
  inc_nrnb(nrnb,eNR_MCLJ,nlj);
  return e;
}	/* TOTAL 11 or 27	*/

real elj_i(int i, rvec x[], t_mc *mc, t_pbc *pbc, 
	   t_forcerec *fr, t_mdatoms *mdatoms, t_nrnb *nrnb)
/* assumes 3 exclusions, indices in sequence, minimum C4 tails */ 
{
  real e=0;
  int j;

#if TIMER
  clock_t start,stop;
  start = clock();
#endif
  for (j=0; j<mdatoms->nr; j++)
    if (((i-j>3)||(i<j))&&(mdatoms->typeA[j]!=mc->gtype))
      e += elj_i_j(i,j,x,pbc,fr,mdatoms,nrnb);
#if TIMER
  stop = clock();
  stopwatch[T_LJ] += stop - start;
#endif

  inc_nrnb(nrnb,eNR_MCLJSEARCH,mdatoms->nr);
  return e;
}

void turn_atom_off(int i, t_mc *mc, t_atoms *atoms, t_mdatoms *mdatoms)
{
  atoms->atom[i].type 	= mdatoms->typeA[i]	= mc->gtype;
  atoms->atom[i].q	= mdatoms->chargeA[i]	= 0.0;
/*
 * Turning the mass off would have some advantages
 * but so far this crashes the simulation after a while
 *
  top->atoms.atom[i].m    = 0.0;
  mdatoms->massA[i]       = 0.0;
  mdatoms->massT[i]       = 0.0;
  mdatoms->invmass[i]     = ALMOST_ZERO;
  mdatoms->tmass         -= mdatoms->massB[i];
*/
}

void turn_atom_on(int i, t_mc *mc, t_atoms *atoms, t_mdatoms *mdatoms)
{
  atoms->atom[i].type 	= mdatoms->typeA[i]	= mdatoms->typeB[i];
  atoms->atom[i].q	= mdatoms->chargeA[i]	= mdatoms->chargeB[i];
}

void turn_group_off(int grpnr, t_mc *mc, t_topology *top, t_mdatoms *mdatoms)
{
  int i;
  fprintf(mclog,"Turning off group %s\n",RESNAME(grpnr));fflush(mclog);
  for (i=0; i<top->atoms.nr; i++)
    if (top->atoms.atom[i].grpnr[egcTC] == grpnr)
      turn_atom_off(i,mc,&(top->atoms),mdatoms);
}

void toggle_m(int m, t_mc *mc, t_topology *top, t_mdatoms *mdatoms)
{
  int i,ai,lipid;

  lipid = ALIVE(m);
  ai = mc->first[lipid][m];
  for (i=0; i<mc->size[lipid]; i++,ai++)
    turn_atom_off(ai, mc, &(top->atoms), mdatoms);

  lipid = !lipid;
  ai = mc->first[lipid][m];
  for (i=0; i<mc->size[lipid]; i++,ai++) 
    turn_atom_on(ai, mc, &(top->atoms), mdatoms);
}

void mk_bond_lookup(t_topology *top, t_mdatoms *md, t_forcerec *fr,
                    t_inputrec *ir, t_groups *grps, t_mc *mc)
{
  int f,ftype,nbonds,t,l,ai,gwnr,ilnr,i,j,last_atom,type;
  int m = 0;  /* first molecule as template */

  for (f=0; f<NFNB; f++) {
    ftype = mc_ftypes[f];
    nbonds = NRAL(ftype) + 1;
    ilnr = top->idef.il[ftype].nr;

    for (l=LIPA; l<NLIP; l++) for (t=T1; t<NTAIL; t++) {
      ai = mc->first[l][m] + mc->t[t].node[l] + 1;
      gwnr = mc->t[t].nr[l] - mc->t[t].nsame;
      snew(mc->t[t].il[l][f].type,gwnr);

//      printf("Okay, I am starting with mol %d tail %d lipid %d ftype %d nbonds %d atom %d grow %d\n",
//             m,t,l,ftype,nbonds,ai,gwnr);fflush(NULL);

      for (i=0; i<gwnr; i++,ai++) {
        j = last_atom = 0;
        while ((ai != last_atom) && (j < ilnr)) {
          type      = top->idef.il[ftype].iatoms[j];
          last_atom = top->idef.il[ftype].iatoms[j+nbonds-1];
//          printf("Okay now I'm on growth %d atom %d j %d type %d last_atom %d\n",
//                  i,ai,j,type,last_atom);fflush(NULL);
          j += nbonds;
        }
        if (j >= ilnr) type = NULL;
//        printf("The type I identified is %d\n",type);fflush(NULL);
        mc->t[t].il[l][f].type[i] = type;
      } /* grow */
    } /* tail */
  } /* ftype */
}

real bonded_energy(int ftype, int natoms, t_iatom *iatoms, t_iparams *iparams, 
               t_inputrec *ir, t_mdatoms *md, t_forcerec *fr, rvec x[], 
               t_pbc *pbc, t_nrnb *nrnb)
{
  real e;
  real dummy,dummy_dvdl;
  rvec dummy_fshift[SHIFTS];
  int ngrp,max,k;

  if (iatoms[0]) {
    if (ftype != F_LJ14) {
      e = interaction_function[ftype].ifunc(natoms+1,iatoms,iparams,(const rvec*)x,
                            NULL,NULL,pbc,NULL,1,&dummy_dvdl,NULL,NULL);
    } else {
      ngrp = ir->opts.ngener;
      max = GID((ngrp-1),(ngrp-1),ngrp) + 1;
      real egrp[max],dummy_e[max];
      for(k=0; k<max; k++) egrp[k] = 0.0;
      dummy = do_nonbonded14(natoms+1,iatoms,iparams,(const rvec*)x,NULL,dummy_fshift,
                             NULL,NULL,1,NULL,md,fr,ngrp,egrp,dummy_e);
      for (k=0,e=0.0; k<max; k++) e += egrp[k];
    }
    if (ftype == F_ANGLES)      inc_nrnb(nrnb,eNR_MCANGLES,1);
    else if (ftype == F_PDIHS)  inc_nrnb(nrnb,eNR_MCPROPER,1);
    else if (ftype == F_IDIHS)  inc_nrnb(nrnb,eNR_MCIMPROPER,1);
    else if (ftype == F_RBDIHS) inc_nrnb(nrnb,eNR_MCRB,1);
    else if (ftype == F_LJ14)   inc_nrnb(nrnb,eNR_MC14,1);
    else gmx_fatal(FARGS,"Force type %d not implemented\n",ftype);
  }

  return e;
}

real de_bonds2(int t, int l, atom_id ai, int i, t_mc *mc, 
               t_iparams *iparams, t_inputrec *ir, t_mdatoms *md,
               t_forcerec *fr, rvec x[], t_pbc *pbc, t_nrnb *nrnb)
{
  int f,ftype,natoms,type,j;
  t_iatom iatoms[10];
  real e=0.0;
#if TIMER
  clock_t start,stop;
  start = clock();
#endif

  for (f=1; f<NFNB; f++) {
    ftype = mc_ftypes[f];
    natoms = NRAL(ftype);
    type = mc->t[t].il[l][f].type[i];
    iatoms[0] = type;
    if (ftype != F_LJ14)
      for (j=1; j<=natoms; j++)
        iatoms[j] = ai - (natoms - j);
    else {
      iatoms[1] = ai - 3;
      iatoms[2] = ai;
    }
    e += bonded_energy(ftype,natoms,iatoms,iparams,ir,md,fr,x,pbc,nrnb);
  }/*f*/

#if TIMER
  stop = clock();
  stopwatch[T_DEBONDS] += stop - start;
#endif
  return e;
}

real de_bonds(int ftype, atom_id an, t_topology *top, t_pbc *pbc, 
                  t_mdatoms *md, rvec x[], t_forcerec *fr, 
                  t_inputrec *ir, t_groups *grps, t_nrnb *nrnb)
/* Incremental 'ftype' bonded energy due to atom 'an' */
{
  real dvdl;
  int nbonds=NRAL(ftype)+1;
  t_iatom iatoms[nbonds];
  int i,j,k;
  real de=0,dummy;
#if TIMER
  clock_t start1,start2,stop1,stop2;
  start1 = start2 = clock();
#endif

  for (i=0; (i<top->idef.il[ftype].nr); ) {  /* This is inefficient (reverse lookup problem) */
    for (j=0; j<nbonds; j++)
      iatoms[j] = top->idef.il[ftype].iatoms[i++];
    if (iatoms[j-1] == an) {
#if TIMER
      stop2 = clock();
      stopwatch[T_BSEARCH] += stop2 - start2;
      start2 = clock();
#endif
      de += bonded_energy(ftype,nbonds-1,iatoms,top->idef.iparams,ir,md,fr,x,pbc,nrnb);
/*    if (ftype != F_LJ14) {
 *      de += interaction_function[ftype].ifunc(nbonds,iatoms,
 *             top->idef.iparams, (const rvec*)x,NULL,NULL,
 *             pbc,NULL,1,&dvdl,NULL,NULL);
 *    } else {
 *      rvec fshift[SHIFTS];
 *      int ngrp = ir->opts.ngener;
 *      int max = GID((ngrp-1),(ngrp-1),ngrp) + 1;
 *      real elj14[max],ecoul14[max];
 *      for(k=0; k<max; k++) elj14[k] = ecoul14[k] = 0.0;
 *      dummy = do_nonbonded14(nbonds,iatoms,top->idef.iparams,
 *                            (const rvec*)x,NULL,fshift,
 *                            NULL,NULL,1,NULL,
 *                            md,fr,ngrp,elj14,ecoul14);
 *      for (k=0; k<max; k++) de += elj14[k];
 *    }*/
#if TIMER
      stop2 = clock();
      stopwatch[T_BOND] += stop2 - start2;
      start2 = clock();
#endif
    }
  }
#if TIMER
  stop1 = stop2 = clock();
  stopwatch[T_BSEARCH] += stop2 - start2;
  stopwatch[T_DEBONDS] += stop1 - start1;
#endif
  return de;
}

real bias_x(atom_id ai, int t, int l, int i, t_mc *mc, t_topology *top, 
              t_pbc *pbc, t_mdatoms *md, rvec x[], t_forcerec *fr, 
              t_inputrec *ir, t_groups *grps, t_nrnb *nrnb)
{
/* Selects a new atom position with intra-molecular energy bias
 * Requires 93 attempts on average (stdev>100)
 */
  int type;
  real e,blen,rand=1,boltz=0;
  rvec xb;
  int ntry=0;

  type   = mc->t[t].il[l][0].type[i];
  blen   = top->idef.iparams[type].shake.dA;

//  printf("bias_x: seed %d ",mc->seed);
  while (rand > boltz) {
    ntry++;
    ran_x(blen,mc->rng,xb);	/* 109	*/
    rvec_add(xb,x[ai-1],x[ai]);	/* 3	*/
#if 0
    e  = de_bonds(F_ANGLES,ai,top,pbc,md,x,fr,ir,grps,nrnb);	/* 163	*/
    e += de_bonds(F_RBDIHS,ai,top,pbc,md,x,fr,ir,grps,nrnb);	/* 247	*/
    e += de_bonds(F_PDIHS,ai,top,pbc,md,x,fr,ir,grps,nrnb);	/* 229 or 0	*/
    e += de_bonds(F_IDIHS,ai,top,pbc,md,x,fr,ir,grps,nrnb);	/* 208 or 0	*/
    e += de_bonds(F_LJ14,ai,top,pbc,md,x,fr,ir,grps,nrnb);	/* 90 or 0	*/
#endif
    e = de_bonds2(t,l,ai,i,mc,top->idef.iparams,ir,md,fr,x,pbc,nrnb);
						/* 846 per dbl-bond	*/
//    rand  = rando(&(mc->seed));	/* 10	*/
    rand = gsl_rng_uniform(mc->rng);	/* 10 */
    boltz = exp(-mc->beta*e);	/* 12	*/
  }
//  printf("%d ntry %d\n",mc->seed,ntry);fflush(NULL);
  inc_nrnb(nrnb,eNR_MCBIAS,ntry);
  return e;
}	/*	TOTAL 50592 or 89931*/

int pick_k(real p[], t_mc *mc)
{
  int k=-1;
  real pcum=0;
// real rand = rando(&(mc->seed));
  real rand = gsl_rng_uniform(mc->rng);
  while (rand > pcum) {
    k++;
    pcum += p[k];
  }
  return k;
}

void calc_nrdf1(t_atoms *atoms,t_idef *idef,t_grpopts *opts,
		      char **gnames,int nstcomm,int comm_mode)
/*
 * Does full blown degrees of freedom calculation (any nr of species)
 * Copied from kernel/readir.c and adjusted for ghosts (it is not possible 
 * to call calc_nrdf from here since it is in the kernel rather than a library)
 */ 
{
  int     ai,aj,i,j,d,g,imin,jmin;
  t_iatom *ia;
  int     *nrdf,*na_vcm,na_tot;
  double  *nrdf_vcm,nrdf_uc,n_sub;

#if TIMER
  clock_t start,stop;
  start = clock();
#endif
  /* Calculate nrdf. 
   * First calc 3xnr-atoms for each group
   * then subtract half a degree of freedom for each constraint
   *
   * Only atoms and nuclei contribute to the degrees of freedom...
   */

  snew(nrdf_vcm,atoms->grps[egcVCM].nr);
  snew(na_vcm,atoms->grps[egcVCM].nr);
  
  for(i=0; i<atoms->grps[egcTC].nr; i++)
    opts->nrdf[i] = 0;
  for(i=0; i<atoms->grps[egcVCM].nr; i++)
    nrdf_vcm[i] = 0;

  snew(nrdf,atoms->nr);
  for(i=0; i<atoms->nr; i++) {
    nrdf[i] = 0;
    if (((atoms->atom[i].ptype == eptAtom) ||
	 (atoms->atom[i].ptype == eptNucleus)) &&
        (atoms->atom[i].type == atoms->atom[i].typeB)) {
      g = atoms->atom[i].grpnr[egcFREEZE];
      /* Double count nrdf for particle i */
      for(d=0; d<DIM; d++)
	if (opts->nFreeze[g][d] == 0)
	  nrdf[i] += 2; 
      opts->nrdf[atoms->atom[i].grpnr[egcTC]] += 0.5*nrdf[i];
      nrdf_vcm[atoms->atom[i].grpnr[egcVCM]] += 0.5*nrdf[i];
    }
  }
  ia=idef->il[F_SHAKE].iatoms;
  for(i=0; (i<idef->il[F_SHAKE].nr); ) {
    /* Subtract degrees of freedom for the constraints,
     * if the particles still have degrees of freedom left.
     * If one of the particles is a vsite or a shell, then all
     * constraint motion will go there, but since they do not
     * contribute to the constraints the degrees of freedom do not
     * change.
     */
    ai=ia[1];
    aj=ia[2];

    if (
        ((atoms->atom[ai].ptype == eptNucleus) || (atoms->atom[ai].ptype == eptAtom)) &&
	((atoms->atom[aj].ptype == eptNucleus) || (atoms->atom[aj].ptype == eptAtom)) &&
        (atoms->atom[ai].type == atoms->atom[ai].typeB) &&
        (atoms->atom[aj].type == atoms->atom[aj].typeB)
       ) {
      if (nrdf[ai] > 0) 
	jmin = 1;
      else
	jmin = 2;
      if (nrdf[aj] > 0)
	imin = 1;
      else
	imin = 2;
      imin = min(imin,nrdf[ai]);
      jmin = min(jmin,nrdf[aj]);
      nrdf[ai] -= imin;
      nrdf[aj] -= jmin;
      opts->nrdf[atoms->atom[ai].grpnr[egcTC]] -= 0.5*imin;
      opts->nrdf[atoms->atom[aj].grpnr[egcTC]] -= 0.5*jmin;
      nrdf_vcm[atoms->atom[ai].grpnr[egcVCM]] -= 0.5*imin;
      nrdf_vcm[atoms->atom[aj].grpnr[egcVCM]] -= 0.5*jmin;
    }
    ia += interaction_function[F_SHAKE].nratoms+1;
    i  += interaction_function[F_SHAKE].nratoms+1;
  }
  ia=idef->il[F_SETTLE].iatoms;
  for(i=0; i<idef->il[F_SETTLE].nr; ) {
    for(ai=ia[1]; ai<ia[1]+3; ai++) {
      opts->nrdf[atoms->atom[ai].grpnr[egcTC]] -= 1;
      nrdf_vcm[atoms->atom[ai].grpnr[egcVCM]] -= 1;
    }
    ia+=2;
    i+=2;
  }

  if (nstcomm != 0) {
    /* Subtract 3 from the number of degrees of freedom in each vcm group
     * when com translation is removed and 6 when rotation is removed
     * as well.
     */
    switch (comm_mode) {
    case ecmLINEAR:
      n_sub = 3;
      break;
    case ecmANGULAR:
      n_sub = 6;
      break;
    default:
      n_sub = 0;
      gmx_incons("Checking comm_mode");
    }
    
    for(i=0; i<atoms->grps[egcTC].nr; i++) {
      /* Count the number of atoms of TC group i for every VCM group */
      for(j=0; j<atoms->grps[egcVCM].nr; j++)
	na_vcm[j] = 0;
      na_tot = 0;
      for(ai=0; ai<atoms->nr; ai++)
	if (atoms->atom[ai].grpnr[egcTC] == i) {
	  na_vcm[atoms->atom[ai].grpnr[egcVCM]]++;
	  na_tot++;
	}
      /* Correct for VCM removal according to the fraction of each VCM
       * group present in this TC group.
       */
      nrdf_uc = opts->nrdf[i];
      opts->nrdf[i] = 0;
      for(j=0; j<atoms->grps[egcVCM].nr; j++)
	if (nrdf_vcm[j] > n_sub) {
	  opts->nrdf[i] += nrdf_uc*((double)na_vcm[j]/(double)na_tot)*
	    (nrdf_vcm[j] - n_sub)/nrdf_vcm[j];
	}
    }
  }

#if DEBUG
  for(i=0; (i<atoms->grps[egcTC].nr); i++) {
    if (opts->nrdf[i] < 0)
      opts->nrdf[i] = 0;
    fprintf(mclog,
	    "Number of degrees of freedom in T-Coupling group %s is %.2f\n",
	    gnames[atoms->grps[egcTC].nm_ind[i]],opts->nrdf[i]);
  }
#endif
  
  sfree(nrdf);
  sfree(nrdf_vcm);
  sfree(na_vcm);
#if TIMER
  stop = clock();
  stopwatch[T_NRDF] += stop - start;
#endif
}

void calc_nrdf2(t_mc *mc,t_grpopts *opts,t_topology *top)
/* Calculates degrees of freedom, only works for ternary LIPA-LIPB-SOL systems */
{
  int i;
  real nrA,nrB,fac;
  static real nrW;
  static bool bFirst=TRUE;
  if (bFirst) {
    for (i=0; i<top->atoms.nres; i++)
      if (strcmp("SOL",*(top->atoms.resname[i]))==0)
        nrW+=6; /* assuming 6 degrees of freedom per water molecule */
    bFirst=FALSE;
  }
  nrA = mc->nr[LIPA]*(2*mc->size[LIPA]+1); /* assume there are (size-1) constraints per lipid */
  nrB = mc->nr[LIPB]*(2*mc->size[LIPB]+1);
  fac = 1.0 - 3.0/(nrA + nrB + nrW);
  opts->nrdf[LIPA] = nrA*fac;
  opts->nrdf[LIPB] = nrB*fac;
  opts->nrdf[2]  = nrW*fac;
  printf("calc_nrdf2: %f %f %f\n",opts->nrdf[LIPA],opts->nrdf[LIPB],opts->nrdf[NLIP]);fflush(NULL);

}

void print_cg(t_block *cgs)
{
  int cgi,k0,k1,k,atom;
  for (cgi=0; cgi<cgs->nr; cgi++) {  /* cycle through all groups */
    k0 = cgs->index[cgi];
    k1 = cgs->index[cgi+1];
    for (k=k0; k<k1; k++) {
      atom = cgs->a[k];
      printf("cgnr %d index %d atom %d\n",cgi,k,atom);
    }
  }
  fflush(NULL);
}


void put_back_in_box(int m, int l, t_mc *mc, t_block *cgs, matrix box, rvec x[], rvec cg_cm[])
{
/* This part is complicated by the need to update the charge group coordinates cg_cm */
  int t,dt,ai,i,d,atom,index;
  int cgi,k0,k1,k,cgai;

  for (t=0; t<NTAIL; t++) {
    dt = mc->t[t].nr[l] - mc->t[t].nsame;
    ai = mc->first[l][m] + mc->t[t].node[l] + 1;
    /* Locate start of corresponding charge groups
       This is inefficient but not part of the mc kernel  */
    for (cgi=0; cgi<cgs->nr; cgi++) {
      k0 = cgs->index[cgi];
      k1 = cgs->index[cgi+1];
      for (k=k0; k<k1; k++)
        if (cgs->a[k] == ai)
          cgai = cgi;
    }
    for (i=0; i<dt; i++,ai++,cgai++) {
//      printf("in_box %d: %.3f %.3f %.3f ",ai,x[ai][XX],x[ai][YY],x[ai][ZZ]);
      for (d=0; d<DIM; d++) {	/* 6	*/
        while (x[ai][d] < 0)          x[ai][d] += box[d][d];
        while (x[ai][d] >= box[d][d]) x[ai][d] -= box[d][d];
        cg_cm[cgai][d] = x[ai][d];
      }
//      printf("--> %.3f %.3f %.3f\n",x[ai][XX],x[ai][YY],x[ai][ZZ]); fflush(NULL);
    }
    ai--; cgai--;
    if (dt > 0 && cgs->a[cgs->index[cgai]] != ai) {
      printf("Charge group reverse lookup discrepancy atom %d cgs->a[%d] %d\n",
                ai+1,cgs->index[cgai],cgs->a[cgs->index[cgai]]);
      for (cgi=0; cgi<cgs->nr; cgi++) {
        k0 = cgs->index[cgi];
        k1 = cgs->index[cgi+1];
        for (k=k0; k<k1; k++)
          if (cgs->a[k] == ai)
            cgai = cgi;
      }
      for (i=0; i<dt; i++,ai++,cgai++) {
        index = cgs->index[cgai];
        atom  = cgs->a[index];
        printf("Tail %d atom %d cg %d cgs->index %d cgs->atom %d\n",t,ai,cgai,index,atom);
      }
      fflush(NULL);
      print_cg(cgs);
      gmx_fatal(FARGS,"Charge group reverse lookup discrepancy atom %d\n", ai+1);
    }
  }
}

void fix_ishifts(int m, int l, t_mc *mc, t_graph *g, matrix box, rvec x[])
/* Based on gmxlib/mshift.c:mk_1shift() and thus should be run only	*
 * on a molecule that is in 'inbox' frame of reference			*/
{
  int t,i,ai,aj,dt,d;
  rvec dx,hbox;

  for (d=0; d<DIM; d++) 
    hbox[d] = 0.5*box[d][d];

  for (t=0; t<NTAIL; t++) {
    dt = mc->t[t].nr[l] - mc->t[t].nsame;
    ai = mc->first[l][m] + mc->t[t].node[l];
//    printf("fix_ishifts: %d (%.2f %.2f %.2f) (%d %d %d)\n",ai+1,x[ai][XX],x[ai][YY],x[ai][ZZ],g->ishift[ai][XX],g->ishift[ai][YY],g->ishift[ai][ZZ]);
    for (i=0; i<dt; i++,ai++) {
      aj = ai + 1;
      rvec_sub(x[ai],x[aj],dx);
//    printf("fix_ishifts: %d (%.2f %.2f %.2f) (%d %d %d) -> ",aj+1,x[aj][XX],x[aj][YY],x[aj][ZZ],g->ishift[aj][XX],g->ishift[aj][YY],g->ishift[aj][ZZ]);
      for (d=0; d<DIM; d++) {

        if (dx[d] < -hbox[d])
          g->ishift[aj][d] = g->ishift[ai][d]-1;
        else if (dx[d] >= hbox[d])
          g->ishift[aj][d] = g->ishift[ai][d]+1;
        else
          g->ishift[aj][d] = g->ishift[ai][d];

      }
//    printf("(%d %d %d)\n",g->ishift[aj][XX],g->ishift[aj][YY],g->ishift[aj][ZZ]); fflush(NULL);
    }
  }
}


/**********************************************************************************
 *
 *	mutate	
 *
 *      Called from:	do_mc
 *      Requires:	elj_i, bias_x, de_bonds2, toggle_m, gsl_rng_uniform
 *			make_vel, calc_nrdf1, shadow, bias_x
 *
 *	Configurational bias mutation is attempted on pre-chosen lipid.
 *	Returns TRUE if mutation is successful.
 *
 **********************************************************************************/

bool mutate(int step, int m, t_mc *mc, t_topology *top, t_mdatoms *mdatoms,
            t_inputrec *ir, t_forcerec *fr, t_pbc *pbc, t_graph *graph, 
            rvec x[], rvec v[], t_commrec *cr, t_groups *grps, t_nrnb *nrnb)
{
  int t,i,j,k,an,ao,dt,new,old,type;
  real e_bon[KMAX],e_nb[KMAX],p[KMAX],norm,enode,W_o,W_n,Pacc;
  rvec xtest[KMAX];
  static int nsucc=0, ntry=0;
  real invkmax = 1.0/mc->kmax;
  bool bAccept=FALSE;
  bool bLog=((step%ir->nstlog==0)&&MASTER(cr));
#if TIMER
  clock_t start,stop;
#endif

  /* Figure out which one is on */
  old = ALIVE(m);
  new = !old;

  if (bLog)
    fprintf(mclog,"\nStep %d Mutating %s, mol %d: atoms (%d-%d)->(%d-%d)\n",
      step,RESNAME(old),m,
      mc->first[old][m],
      mc->first[old][m]+mc->size[old]-1,
      mc->first[new][m],
      mc->first[new][m]+mc->size[new]-1
    );

#if TIMER
  start = start_tot = clock();
#endif
/*********************** REGROW OLD TAILS **************************/
  for (t=0; t<NTAIL; t++) { /* tails and nodes off at start */
    dt = mc->t[t].nr[old] - mc->t[t].nsame + 1;
    ao = mc->first[old][m] + mc->t[t].node[old];
    for (i=0; i<dt; i++,ao++)
      mdatoms->typeA[ao] = mc->gtype;
  }

  for (enode=0,t=0; t<NTAIL; t++) {			/* Node energy 			*/
    ao = mc->first[old][m] + mc->t[t].node[old];	/* only LJ			*/
    mdatoms->typeA[ao] = mdatoms->typeB[ao];		/* since even double bonds	*/
    enode += elj_i(ao,x,mc,pbc,fr,mdatoms,nrnb);	/* have same bonded energies	*/
  }
  W_o = exp(-mc->beta*enode);	/* 12	*/
  if (bLog)
    fprintf(mclog,"Node atoms: e %g W_o %g\n",enode,W_o);
  for (t=0; t<NTAIL; t++) {  /* Regrow old tail */
    dt = mc->t[t].nr[old] - mc->t[t].nsame;
    ao = mc->first[old][m] + mc->t[t].node[old] + 1;
    for (i=0; i<dt; i++,ao++) {
      mdatoms->typeA[ao] = mdatoms->typeB[ao];
      for (norm=0,k=0; k<mc->kmax; k++) {
        e_bon[k] = (k?		/* 60000	*/
                      bias_x(ao,t,old,i,mc,top,pbc,mdatoms,x,fr,ir,grps,nrnb):
                      de_bonds2(t,old,ao,i,mc,top->idef.iparams,ir,mdatoms,fr,x,pbc,nrnb));
/*
                      de_bonds(F_ANGLES,ao,top,pbc,mdatoms,x,fr,ir,grps,nrnb)+
                      de_bonds(F_RBDIHS,ao,top,pbc,mdatoms,x,fr,ir,grps,nrnb)+
                      de_bonds(F_PDIHS,ao,top,pbc,mdatoms,x,fr,ir,grps,nrnb)+
                      de_bonds(F_IDIHS,ao,top,pbc,mdatoms,x,fr,ir,grps,nrnb)+
                      de_bonds(F_LJ14,ao,top,pbc,mdatoms,x,fr,ir,grps,nrnb));
*/

        e_nb[k]  = elj_i(ao,x,mc,pbc,fr,mdatoms,nrnb);	/* 11*N	*/
        p[k]     = exp(-mc->beta*e_nb[k]);		/* 12	*/
        norm    += p[k];				/* 1	*/
        copy_rvec(x[ao],xtest[k]);			/* 3	*/
      }		/* kmax*(60000 + 11*N)	*/
      copy_rvec(xtest[0],x[ao]);
      W_o *= invkmax*norm;
      if (bLog)
        fprintf(mclog,"Tail%d-%d e[%d]={ %10.6f %11.6f } W_o %13g\n",t,i,0,e_bon[0],e_nb[0],W_o);
    }
  }

  toggle_m(m,mc,top,mdatoms);
#if TIMER
  stop = clock();
  stopwatch[T_OLD] += stop - start;
  start = clock();
#endif

/************************ GROW NEW TAILS *************************/
  for (t=0; t<NTAIL; t++) { /* tails and nodes off at start */
    dt = mc->t[t].nr[new] - mc->t[t].nsame + 1;
    an = mc->first[new][m] + mc->t[t].node[new];
    for (i=0; i<dt; i++,an++)
      mdatoms->typeA[an] = mc->gtype;
  }
  for (enode=0,t=0; t<NTAIL; t++) {  /* node energy */
    an = mc->first[new][m] + mc->t[t].node[new];
    mdatoms->typeA[an] = mdatoms->typeB[an];
    enode += elj_i(an,x,mc,pbc,fr,mdatoms,nrnb);
  }
  W_n = exp(-mc->beta*enode);
  if (bLog)
    fprintf(mclog,"Node atoms: e %g W_n %g\n",enode,W_n);
  int nnew = mc->t[T1].nr[new] - mc->t[T1].nsame +
             mc->t[T2].nr[new] - mc->t[T2].nsame;
  atom_id new_atoms[nnew];
  for (t=0,j=0; t<NTAIL; t++) {  /* Grow new tail */
    dt = mc->t[t].nr[new] - mc->t[t].nsame;
    an = mc->first[new][m] + mc->t[t].node[new] + 1;
    for (i=0; i<dt; i++,j++,an++) {
      new_atoms[j] = an;
      mdatoms->typeA[an] = mdatoms->typeB[an];
      for (norm=0,k=0; k<mc->kmax; k++) {
        e_bon[k] = bias_x(an,t,new,i,mc,top,pbc,mdatoms,x,fr,ir,grps,nrnb);
        e_nb[k]  = elj_i(an,x,mc,pbc,fr,mdatoms,nrnb);
        p[k]        = exp(-mc->beta*e_nb[k]);
        norm       += p[k];
        copy_rvec(x[an],xtest[k]);
      }
      for (k=0; k<mc->kmax; k++) p[k]/=norm;
      k = pick_k(p,mc);
      copy_rvec(xtest[k],x[an]);
      W_n *= invkmax*norm;
      if (bLog)
        fprintf(mclog,"Tail%d-%d e[%d]={ %10.6f %11.6f } W_n %13g\n",t,i,k,e_bon[k],e_nb[k],W_n);
    }
  }
#if TIMER
  stop = clock();
  stopwatch[T_NEW] += stop - start;
  start = clock();
#endif

  Pacc = mc->alpha[new]*(W_n/W_o);
  if (bLog)
    fprintf(mclog,"FINAL W_o %g W_n %g Pacc %g\n",W_o,W_n,Pacc);

//  if(rando(&(mc->seed)) < Pacc) { /* accept */
  if(gsl_rng_uniform(mc->rng) < Pacc) { /* accept */
    if (bLog) fprintf(mclog,"ACCEPTED ");
    mc->nr[new]++;
    mc->nr[old]--;
    gsl_mspeed(ir->opts.ref_t[new],2*nnew,nnew,new_atoms,&(top->atoms),v,mc->rng);
//    make_vel(new,m,ir->opts.ref_t[new],mc,&(top->atoms),v,ir);
//    calc_nrdf2(mc,&(ir->opts),top);
    char *gnames[top->atoms.grps[egcNR].nr+1];
    for(i=0;i<top->atoms.grps[egcTC].nr+1;i++) 
      gnames[i] = *(top->atoms.grpname[i]);
    calc_nrdf1(&(top->atoms),&(top->idef),&(ir->opts),gnames,ir->nstcomm,ir->comm_mode);
    bAccept = TRUE;
  } else {                        /* reject */
    toggle_m(m,mc,top,mdatoms);
    if (bLog) fprintf(mclog,"REJECTED ");
  }
  i=countB(mc,top);
  nsucc+=bAccept; ntry++;
  if (bLog) {
    fprintf(mclog,"Success-Rate %d/%d %.2f%% Composition %s %d %s %d\n",
        nsucc,ntry,100*(real)nsucc/(real)ntry,RESNAME(LIPA),mc->nmol-i,RESNAME(LIPB),i);
    fflush(mclog);
  }

#if TIMER
  stop = clock();
  stopwatch[T_ACCREJ] += stop - start;
#endif

  return bAccept;
}	/* (sizeA+sizeB-2*nsame)*kmax*(60000+11*N)+22*N	*/



/**********************************************************************************
 *
 *	init_mc	
 *
 *      Called from:	do_md
 *      Requires:	gsl_rng_*, mk_bond_lookup, turn_atom_off, turn_group_off
 *			calc_nrdf1, shadow, bias_x
 *
 * 	This is called once to read and interpret the input configuration.
 *	It assumes the first two temperature groups correspond to the mutable
 *	lipid pairs. If it is not a continuation run, the second group is
 *      reconfigured and turned off. It also assumes that the topology file
 *	lists the tail atoms sequentially.
 *
 **********************************************************************************/
void init_mc(t_topology *top, t_inputrec *ir, t_commrec *cr,
                t_mdatoms *mdatoms, t_graph *graph, rvec x[], 
                rvec v[], matrix box, t_mc *mc, 
		t_forcerec *fr, t_groups *grps, t_nrnb *nrnb)
{
  int i,j,n,ai,an1[NLIP],an2[NLIP],last,tail,nLH1,nn[NLIP];
  int natom[NLIP],nmol[NLIP],grpnr,resnr,last_resnr,ngrp;
  int l,start,nr,index,type,f;
  real blen;

  mclog = ffopen("mc.log","w");
  fprintf(mclog,"--------------------------------------------------\nInitializing MC!\n");

  /* load input parameters */
  rc2 = sqr(fr->rvdw);
  mc->kmax  = ir->mckmax;
  mc->alpha[LIPA] = ir->mcactivityratio;
  mc->alpha[LIPB] = 1.0/mc->alpha[LIPA];
  if (ir->opts.ref_t[LIPA] == ir->opts.ref_t[LIPB])
    mc->beta  = 1.0/(KB*ir->opts.ref_t[LIPA]);
  else
    gmx_fatal(FARGS,"MC: Unequal group temperatures %f %f\n",
              ir->opts.ref_t[LIPA],ir->opts.ref_t[LIPB]);

  /* Sync random number generator */
  gsl_rng_env_setup();
  mc->rng = gsl_rng_alloc (gsl_rng_default);
  if (ir->mcseed == 0) {
    FILE *RNG_IN = fopen ("rng","r");
    if (RNG_IN == NULL) {
      fprintf(mclog,"Using default random seed %d\n",gsl_rng_default_seed);
      gsl_rng_set(mc->rng, gsl_rng_default_seed);
    } else {
      fprintf(mclog,"Continuing from last random seed\n");
      gsl_rng_fread(RNG_IN, mc->rng);
      fclose(RNG_IN);
    }
  } else {
    fprintf(mclog,"Using random seed %d\n",ir->mcseed);
    gsl_rng_set (mc->rng, ir->mcseed);
  }

#if 0  /* This is the old rng, can be retired */
#ifdef GMX_MPI
  if (MASTER(cr)) { 
    mc->seed = make_seed(); /* determine random seed on master node */
    gmx_txs(cr->right, &(mc->seed), sizeof(mc->seed)); /* send right */
  }
  else {
    if (cr->nodeid < cr->nnodes - 1) {
      gmx_rxs(cr->left,  &(mc->seed), sizeof(mc->seed)); /* get from left */
      gmx_txs(cr->right, &(mc->seed), sizeof(mc->seed)); /* send right */
    }
    else
      gmx_rxs(cr->left,  &(mc->seed), sizeof(mc->seed)); /* last one */
  }
#else
  mc->seed = make_seed();
#endif
  mc->seed = 198742981; // debug
  fprintf(mclog,"MC seed %d\n",mc->seed);
#endif

  /* Clone HW type for ghosts */
  for (i=0; (strcmp("HW",*(top->atoms.atomtype[i]))!=0) && (i<top->atoms.nr); i++) {}
  if(MASTER(cr)) {
    fprintf(mclog,"Cloning atom %d (type=%d, name=%s) for mc ghost atoms\n",
           i,mc->gtype,*(top->atoms.atomtype[i]));
  }
  if (i<top->atoms.nr) 
    mc->gtype = top->atoms.atom[i].type;
  else 
    gmx_fatal(FARGS,"MC: No HW atom found for use as ghost atom type\n");

  /* Determine lipid indices using temperature groups          *
   * so it will work even if molecules are sorted and shuffled */
  for (n=0;n<2;n++) {
    last_resnr = -1;
    nmol[LIPA]  = nmol[LIPB]  = 0;
    natom[LIPA] = natom[LIPB] = 0;
    if (n) {
      snew(mc->first[LIPA],mc->nmol);
      snew(mc->first[LIPB],mc->nmol);
    }
    for (ai=0; ai<top->atoms.nr; ai++) {
      resnr = top->atoms.atom[ai].resnr;
      grpnr = top->atoms.atom[ai].grpnr[egcTC];
      if (grpnr < NLIP) { /* mutable lipids must be first two groups */
        natom[grpnr]++;
        if (resnr != last_resnr) {
          if (n) mc->first[grpnr][nmol[grpnr]] = ai;
          nmol[grpnr]++;
        }
      }
      last_resnr = resnr;
    }
    if (!n) {
      if (nmol[LIPA] != nmol[LIPB])
        gmx_fatal(FARGS,"MC: Lipid pair mismatch %d %d\n",
          nmol[LIPA],nmol[LIPB]);
      mc->nmol = nmol[LIPA];
    }
  }
  mc->size[LIPA] = natom[LIPA]/mc->nmol;
  mc->size[LIPB] = natom[LIPB]/mc->nmol;

  /* Determine tail beginnings and endings with help of LP3's	*
   * thus only works for ITP files that order tail-1 atoms and	*
   * tail-2 atoms sequentially.					*/
  for (l=LIPA; l<NLIP; l++) {
    j = mc->first[l][0];
    for (ai=j; (strcmp(*(top->atoms.atomtype[ai]),"LP3")!=0); ai++) {
      if (ai == j+mc->size[l])
        gmx_fatal(FARGS,"MC: Did not find tail endpoint (looking for LP3)\n");
    } 
    last = ai - j;
    mc->t[T1].nr[l]    = last+1;
    mc->t[T2].nr[l]    = mc->size[l] - mc->t[T1].nr[l];
    mc->t[T1].first[l] = 0;
    mc->t[T2].first[l] = last+1;
    mc->t[T1].last[l]  = last;
    mc->t[T2].last[l]  = mc->size[l]-1;
  }

  /* Find tail growth nodes */
  bool bSame;
  int nsame, atomA, atomB, lastA, lastB;
  for (tail=T1; tail<NTAIL; tail++) {
    bSame = TRUE;
    nsame = 0;
    atomA = mc->first[LIPA][0] + mc->t[tail].first[LIPA] - 1;
    atomB = mc->first[LIPB][0] + mc->t[tail].first[LIPB] - 1;
    lastA = atomA + mc->t[tail].nr[LIPA];
    lastB = atomB + mc->t[tail].nr[LIPB];
    while (bSame) {
      atomA++; atomB++; nsame++;
      bSame = (top->atoms.atom[atomA].type == top->atoms.atom[atomB].type) &
              (atomA < lastA) & (atomB < lastB) ; 
    }
    mc->t[tail].nsame = nsame;
    mc->t[tail].node[LIPA] = mc->t[tail].first[LIPA] + nsame - 1;
    mc->t[tail].node[LIPB] = mc->t[tail].first[LIPB] + nsame - 1;
  }

  /* Assume that 'LH1' terminates the headgroups and check that lipids	*
   * do share the same headgroup					*/
  atomA = mc->first[LIPA][0] + mc->t[T1].first[LIPA];
  lastA = atomA + mc->t[T1].nr[LIPA] - 1;
  while ((strcmp(*(top->atoms.atomtype[atomA]),"LH1") != 0) && (atomA <= lastA)) atomA++;
  if (mc->t[T1].nsame < (atomA - mc->first[LIPA][0] + 1))
    gmx_fatal(FARGS,"Immutable lipids - these headgroups are not the same\n");
  if (mc->t[T2].nsame < 3)
    gmx_fatal(FARGS,"Tail 2 must have at least 3 common atoms\n");
    
  /* Record bonded force types for growth atom. It assumes all lipids	*
   * are identical in use of first lipid as template.			*/
  mk_bond_lookup(top,mdatoms,fr,ir,grps,mc);

  /* Messages */
  if(MASTER(cr)) {
    fprintf(mclog,"Found activity ratio %s/%s= %g\n",
           RESNAME(LIPA),RESNAME(LIPB),mc->alpha[LIPA]);
    fprintf(mclog,"Found temperature %gK, beta=%g\n",ir->opts.ref_t[LIPA],mc->beta);
    fprintf(mclog,"Found %d lipid pairs. \nLipid sizes %s %d %s %d\n",
           mc->nmol,RESNAME(LIPA),mc->size[LIPA],RESNAME(LIPB),mc->size[LIPB]);
    fprintf(mclog,"Common atoms (nr=%d,%d) will be position/velocity mapped:\n",
           mc->t[T1].nsame,mc->t[T2].nsame);
    for (l=LIPA; l<NLIP; l++)
      fprintf(mclog,"   %s: (%d-%d,%d-%d)\n",RESNAME(l),
        mc->t[T1].first[l],mc->t[T1].node[l],
        mc->t[T2].first[l],mc->t[T2].node[l]);

    for (l=LIPA; l<NLIP; l++) {
      fprintf(mclog,"\n%s Growth Table:\ni  btype   blen  angle   pdih   idih  rbdih   lj14\n",RESNAME(l));
      for (tail=T1; tail<NTAIL; tail++) {
        nr = mc->t[tail].nr[l] - mc->t[tail].nsame;
        for (i=0; i<nr; i++) {
          type = mc->t[tail].il[l][0].type[i];
          blen = top->idef.iparams[type].shake.dA;
          fprintf(mclog,"%d %6d %6.3f ",i,type,blen);
          for (f=1; f<NFNB; f++) {
            type = mc->t[tail].il[l][f].type[i];
            fprintf(mclog,"%6d ",type);
          }
          fprintf(mclog,"\n");
        }
      }
    }

  }

  /* Is it a continuation run or do we need to ghostify it? */
  int ghost_atom_nr=0, ghost_mol_nr=0, last_mol=-1, mol;
  for (ai=0; ai<top->atoms.nr; ai++) {
    mol = top->atoms.atom[ai].resnr;
    if (x[ai][XX] > 0.8*GHOST_SHIFTX) {
      ghost_atom_nr++;
      if (mol != last_mol) ghost_mol_nr++;
      turn_atom_off(ai,mc,&(top->atoms),mdatoms);
      x[ai][XX]-=GHOST_SHIFTX;
    }
    last_mol = mol;
  }

  if (ghost_mol_nr == 0)
    turn_group_off(LIPB,mc,top,mdatoms);  /* Turn off all 'B' lipids (2nd energy group) */

  /* Initialize degrees of freedom */
  mc->nr[LIPB] = countB(mc,top);
  mc->nr[LIPA] = mc->nmol - mc->nr[LIPB];
  fprintf(mclog,"\nComposition %s %d %s %d\n",RESNAME(LIPA),mc->nr[LIPA],RESNAME(LIPB),mc->nr[LIPB]);
//  calc_nrdf2(mc,&(ir->opts),top);
  ngrp = top->atoms.grps[egcTC].nr;
  fprintf(mclog,"Degrees of freedom ");
  for(i=0; i<ngrp; i++) fprintf(mclog,"%14f ",ir->opts.nrdf[i]);
  fprintf(mclog," -->\n");
  char *gnames[ngrp+1];
  for(i=0;i<ngrp+1;i++) gnames[i] = *(top->atoms.grpname[i]);
  calc_nrdf1(&(top->atoms),&(top->idef),&(ir->opts),gnames,ir->nstcomm,ir->comm_mode);
  fprintf(mclog,"Degrees of freedom ");
  for(i=0; i<ngrp; i++) fprintf(mclog,"%14f ",ir->opts.nrdf[i]);
  fprintf(mclog,"\n");

  if (ghost_mol_nr == 0) {  /* Map common atoms and regrow excess */
    shadow(mc,top,graph,fr,box,x,v);
    int dt,test=4,k,kmin;
    t_pbc pbc;
    rvec xtest[test];
    real ebon,emin;
    set_pbc(&pbc,box);

    fprintf(stdout,"Initializing %s coordinates (this may take a while)...\nMol ",RESNAME(LIPB));
    fflush(stdout);
    for (mol=0; mol<mc->nmol; mol++) {
      fprintf(stdout,"%d ",mol);fflush(stdout);
      for (tail=0; tail<NTAIL; tail++) {
        dt = mc->t[tail].nr[LIPB] - mc->t[tail].nsame;
        ai = mc->first[LIPB][mol] + mc->t[tail].node[LIPB] + 1;
        for (i=0; i<dt; i++,ai++) {
          for (k=0; k<test; k++) {
            ebon = bias_x(ai,tail,LIPB,i,mc,top,&pbc,mdatoms,x,fr,ir,grps,nrnb);
            copy_rvec(x[ai],xtest[k]);
            if (k==0||ebon<emin) {
              kmin = k;
              emin = ebon;
            }
          }
          copy_rvec(xtest[kmin],x[ai]);
        }
      }
    }
  }
  else if (ghost_mol_nr != mc->nmol)
    gmx_fatal(FARGS,"Coordinates are ambiguous, found %d ghost"
              " atoms %d ghost mols\n",ghost_atom_nr,ghost_mol_nr);

    fprintf(mclog,"--------------------------------------------------\n\n");
    fflush(mclog);

}


void check_energy(t_mc *mc, t_topology *top, t_pbc *pbc,
                  t_mdatoms *md, rvec x[], t_forcerec *fr,
                  t_inputrec *ir, t_groups *grps, t_nrnb *nrnb)
{
  int i;
  real e1,e2,e3,e4,e5;
  e1 = e2 = e3 = e4 = e5 = 0.0;

  for (i=0; i<top->atoms.nr; i++) {
    if (md->typeA[i] == md->typeB[i]) {
      e1 += de_bonds(F_ANGLES,i,top,pbc,md,x,fr,ir,grps,nrnb);
      e2 += de_bonds(F_PDIHS, i,top,pbc,md,x,fr,ir,grps,nrnb);
      e3 += de_bonds(F_RBDIHS,i,top,pbc,md,x,fr,ir,grps,nrnb);
      e4 += de_bonds(F_IDIHS, i,top,pbc,md,x,fr,ir,grps,nrnb);
      e5 += de_bonds(F_LJ14,  i,top,pbc,md,x,fr,ir,grps,nrnb);
    }
  }
  fprintf(mclog,"\ncheck_energy:\n");
  fprintf(mclog,"F_ANGLES %f F_PDIHS %f F_RBDIHS %f "
                "F_IDIHS %f F_LJ14 %f\n",e1,e2,e3,e4,e5);
  fflush(mclog);

}

void check_energy2(t_mc *mc, t_topology *top, t_pbc *pbc,
                  t_mdatoms *md, rvec x[], t_forcerec *fr,
                  t_inputrec *ir, t_groups *grps, t_nrnb *nrnb)
{
  real e[20]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  real e1,e2,e3,e4,e5;
  int l,m,i,ai,real;

  for (l=LIPA; l<NLIP; l++) {
    for (m=0; m<mc->nmol; m++) {
      for (i=0; i<mc->size[l]; i++) {
        ai = mc->first[l][m] + i;
        e1 = de_bonds(F_ANGLES,ai,top,pbc,md,x,fr,ir,grps,nrnb);
        e2 = de_bonds(F_PDIHS, ai,top,pbc,md,x,fr,ir,grps,nrnb);
        e3 = de_bonds(F_RBDIHS,ai,top,pbc,md,x,fr,ir,grps,nrnb);
        e4 = de_bonds(F_IDIHS, ai,top,pbc,md,x,fr,ir,grps,nrnb);
        e5 = de_bonds(F_LJ14,  ai,top,pbc,md,x,fr,ir,grps,nrnb);
        e[l]   += e1;
        e[l+2] += e2; 
        e[l+4] += e3; 
        e[l+6] += e4; 
        e[l+8] += e5; 
        real = (md->typeA[ai] == md->typeB[ai]);
        printf("l %d m %d i %d real %d\n",l,m,i,real);fflush(NULL);
        e[real+10] += e1;
        e[real+12] += e2;
        e[real+14] += e3;
        e[real+16] += e4;
        e[real+18] += e5;
      }
    }
  }
  fprintf(mclog,"\ncheck_energy2:\n");
  fprintf(mclog,"F_ANGLES: %s %9.3f %s %9.3f  real %9.3f ghost %9.3f  tot %9.3f\n",
           RESNAME(LIPA),e[0],RESNAME(LIPB),e[1],e[11],e[10],e[10]+e[11]);
  fprintf(mclog,"F_PDIHS:  %s %9.3f %s %9.3f  real %9.3f ghost %9.3f  tot %9.3f\n",
           RESNAME(LIPA),e[2],RESNAME(LIPB),e[3],e[13],e[12],e[12]+e[13]);
  fprintf(mclog,"F_RBDIHS: %s %9.3f %s %9.3f  real %9.3f ghost %9.3f  tot %9.3f\n",
           RESNAME(LIPA),e[4],RESNAME(LIPB),e[5],e[15],e[14],e[14]+e[15]);
  fprintf(mclog,"F_IDIHS:  %s %9.3f %s %9.3f  real %9.3f ghost %9.3f  tot %9.3f\n",
           RESNAME(LIPA),e[6],RESNAME(LIPB),e[7],e[17],e[16],e[16]+e[17]);
  fprintf(mclog,"F_LJ14:   %s %9.3f %s %9.3f  real %9.3f ghost %9.3f  tot %9.3f\n",
           RESNAME(LIPA),e[8],RESNAME(LIPB),e[9],e[19],e[18],e[18]+e[19]);
  fflush(mclog);

}



/* pin overlapping atoms and shift excess ones */
void shadow(t_mc *mc, t_topology *top, t_graph *graph, t_forcerec *fr, 
            matrix box, rvec x[], rvec v[])
{
  int m,i,j,aa,ag,ga,gg,last,extra,t;
  rvec shift;

  for (m=0; m<mc->nmol; m++) {
    ga = ALIVE(m);
    gg = !ga;
    for (t=0; t<NTAIL; t++) {	/* 3*(dt1+dt2) + 3*nt	*/
      last = mc->t[t].nsame-1;
      extra = mc->t[t].nr[gg] - mc->t[t].nsame;
      aa = mc->first[ga][m] + mc->t[t].first[ga];
      ag = mc->first[gg][m] + mc->t[t].first[gg];
      for (i=0; i<=last; i++,aa++,ag++) {
        if (i==last && extra) {
          rvec_sub(x[aa],x[ag],shift);	/* 3		*/
          for (j=1; j<=extra; j++) {	/* 3*extra	*/
            rvec_add(shift,x[ag+j],x[ag+j]);
          }
        }
        copy_rvec(x[aa],x[ag]);
        copy_rvec(v[aa],v[ag]);
      }
    }
    put_back_in_box(m,gg,mc,&(top->blocks[ebCGS]),box,x,fr->cg_cm);	/* 3	*/
  }
}


void check_inbox(t_topology *top, matrix box, rvec x[], rvec cg_cm[])
{
  int i,d,atom_errors=0,cg_errors=0,k0,k1,k,ak;
  rvec xcg;
  for (i=0; i<top->blocks[ebCGS].nr; i++) {
    k0 = top->blocks[ebCGS].index[i];
    k1 = top->blocks[ebCGS].index[i+1];
    clear_rvec(xcg);
    for (k=k0; k<k1; k++) {
      ak = top->blocks[ebCGS].a[k];
      for (d=0; d<DIM; d++)
        xcg[d] += x[ak][d];
    }
    for (d=0; d<DIM; d++) {
      xcg[d] /= (real)(k1-k0);
      if ((cg_cm[i][d] >= box[d][d]) || (cg_cm[i][d] < 0))
        cg_errors++;
      if ((xcg[d] >= box[d][d]) || (xcg[d] < 0)) {
        atom_errors++;
        for (k=k0; k<k1; k++) {
          ak = top->blocks[ebCGS].a[k];
          printf("Atom %d x %f %f %f\n",ak+1,x[ak][XX],x[ak][YY],x[ak][ZZ]);
        }
        printf("avg  %f %f %f\n",xcg[XX],xcg[YY],xcg[ZZ]);
        printf("cgcm %f %f %f\n",cg_cm[i][XX],cg_cm[i][YY],cg_cm[i][ZZ]);
        printf("box  %f %f %f\n",box[XX][XX],box[YY][YY],box[ZZ][ZZ]);
      }
    }
  }
  if (cg_errors) printf("WARNING ");
  printf("check_inbox: %d errors in charge group coords\n",cg_errors);
  if (atom_errors) printf("WARNING ");
  printf("check_inbox: %d errors in atom coords\n",atom_errors);
  fflush(NULL);
}

void check_whole(t_graph *g, rvec x[])
{
  int i,j,aj;
  rvec dx;
  real dist;
  bool bPass = TRUE;

  for (i=0; i<g->nnodes; i++) {
    for (j=0; j<g->nedge[i]; j++) {
      aj = g->edge[i][j];
      rvec_sub(x[i],x[aj],dx);
      dist = norm(dx);
      if (dist > 1.0) {
        printf("WARNING: There could be a pbc problem.\n"
               "  Atoms %d %d are %.2fnm apart.\n",i+1,aj+1,dist);
        fflush(NULL);
        bPass = FALSE;
      }
    }
  }

  if (bPass) {
    printf("System passed whole molecule check.\n");
    fflush(NULL);
  }
}

real check_dist(int ai, int aj, rvec x[])
{
  rvec dx;
  rvec_sub(x[ai],x[aj],dx);
  return norm(dx);
}

bool do_mc(int step, rvec x[], rvec v[], matrix box, t_forcerec *fr,
           t_topology *top, t_inputrec *ir, t_mdatoms *mdatoms, 
           t_commrec *cr, t_graph *graph, t_mc *mc, t_groups *grps, t_nrnb *nrnb)
{
  int i,m,new;
  t_pbc pbc;
  bool bChange = FALSE, bMutate;
  static bool bFirst = TRUE;
#if TIMER
  clock_t start,stop;
#endif

  if(bFirst){
    bFirst = FALSE;
#if TIMER
    for (i=0;i<T_NR;i++) stopwatch[i] = 0;
#endif
  }

  set_pbc(&pbc,box);
  shadow(mc,top,graph,fr,box,x,v);
//  check_inbox(top, box, x, fr->cg_cm);
  mk_mshift(stderr,graph,box,x);	/* otherwise occasional unwholesome molecules */
  shift_self(graph,box,x);		/* make whole */
//  check_whole(graph,x);

  bMutate = (step%1 == 0);
  if(bMutate) {
//    m = RANMOL(mc); 
    m = gsl_rng_uniform_int(mc->rng, mc->nmol);
    new = !ALIVE(m);
//    printf("\nStep %d Molecule %d\n",step,m);fflush(NULL);
#if TIMER
    start = clock();
#endif
    bChange = mutate(step,m,mc,top,mdatoms,ir,fr,&pbc,graph,x,v,cr,grps,nrnb);
#if TIMER
    start = clock();
    stopwatch[T_MUTATE] += stop - start;
#endif
  }

//  if(step%ir->nstenergy == 0) check_energy(mc,top,&pbc,mdatoms,x,fr,ir,grps,nrnb);

  if(step%ir->nstxout==0 || step == ir->nsteps) {
    write_ndxtraj("traj.ndx",step,&(top->atoms));
    if(step) {
      FILE *rng_out = fopen("rng","w");
      gsl_rng_fwrite(rng_out, mc->rng);
      fclose(rng_out);
    }
  }
//  check_whole(graph,x);
  unshift_self(graph,box,x);
  if (bMutate) {
    put_back_in_box(m,new,mc,&(top->blocks[ebCGS]),box,x,fr->cg_cm);
    fix_ishifts(m,new,mc,graph,box,x);
  }

//  check_inbox(top, box, x, fr->cg_cm);

//  shift_self(graph,box,x);
//  check_whole(graph,x);
//  unshift_self(graph,box,x);


#if TIMER
  if(step%10 == 0) {
    real t[T_NR];
    for (i=0; i<T_NR; i++) t[i] = (real)stopwatch[i]/CLOCKS_PER_SEC*1000;
    printf("TIME(ms)         CUM.       AVG.\n");
//  printf("BSEARCH      %10.1f %6.1f\n",t[T_BSEARCH],t[T_BSEARCH]/(real)step);
//  printf("BOND       %10.1f %6.1f\n",t[T_BOND],t[T_BOND]/(real)step);
    printf("  RANX     %10.1f %6.1f\n",t[T_RANX],t[T_RANX]/(real)step);
    printf("DEBONDS    %10.1f %6.1f\n",t[T_DEBONDS],t[T_DEBONDS]/(real)step);
    printf("LJ         %10.1f %6.1f\n",t[T_LJ],t[T_LJ]/(real)step);
    printf("MUTATE     %10.1f %6.1f\n",t[T_MUTATE],t[T_MUTATE]/(real)step);
    printf("  OLD      %10.1f %6.1f\n",t[T_OLD],t[T_OLD]/(real)step);
    printf("  NEW      %10.1f %6.1f\n",t[T_NEW],t[T_NEW]/(real)step);
    printf("  ACCREJ   %10.1f %6.1f\n",t[T_ACCREJ],t[T_ACCREJ]/(real)step);
    printf("    RANV   %10.1f %6.1f\n",t[T_RANV],t[T_RANV]/(real)step);
    printf("    NRDF   %10.1f %6.1f\n",t[T_NRDF],t[T_NRDF]/(real)step);
    fflush(NULL);
  }
#endif

  return bChange;
}

